/* Copyright (c) 2009-2010 Ondrej Donek.
   See the file LICENSE.txt for licensing information. */

/**
 * Handler for our note column in the main Thunderbird's messages tree.
 * 
 * @todo Sorting is not working!!!
 * @link http://developer.mozilla.org/en/Extensions/Thunderbird/Creating_a_Custom_Column
 * @type Object
 */
var tbNotesColumnHandler = {

	getCellText: function(aRow, aCol) {
		return "";
	},

	getImageSrc: function(aRow, aCol) {
		
		var selMsg = gDBView.getMsgHdrAt(aRow); // returns nsIMsgHdr
		var note   = tbNotes.checkNoteForTargetMsg({
			key     : selMsg.messageKey,
			subject : selMsg.subject,
			id      : selMsg.messageId
		});
		
		return (note === true) 
				? "chrome://tbnotes-icons/content/note-flag14.png" 
				: "chrome://tbnotes-icons/content/note-no-flag14.png";
	},

	isString: function() {
		// This is only for sorting purpose
		return false; 
	},

	getSortStringForRow: function(aHdr) { return ""; },

	getCellProperties: function(aRow, aCol, aProps) {},
	getRowProperties: function(aRow, aProps) {},
	getSortLongForRow: function(aHdr) {
		var selMsg = gDBView.getMsgHdrAt(aRow); // returns nsIMsgHdr
		var note   = tbNotes.checkNoteForTargetMsg({
			key     : selMsg.messageKey,
			subject : selMsg.subject,
			id      : selMsg.messageId
		});
		
		return (note === true) ? 1 : 0;
	}
	
}; // End of tbNotesColumnHandler

/**
 * Implements Components.interfaces.nsIObserver for message "MsgCreateDBView".
 *
 * @type Object
 */
var tbNotesCreateDbObserver = {

	observe: function(aMsgFolder, aTopic, aData) {  
		gDBView.addColumnHandler("tbn-note-treecol", tbNotesColumnHandler);
	}
	
}; // End of tbNotesCreateDbObserver

/**
 * Object implementing listener (will be added to gMessageListeners).
 *
 * @see chrome://messenger/content/msgHdrViewOverlay.js
 * @type Object
 */
var tbNotesMessageListener = {
	
	onStartHeaders: function()
	{
		var XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";
		var selMsg = tbNotes.getSelectedMessageDescription();
		var hbox = document.getElementById("tbn-notesmsgheaderviewpane-hbox");
		//if(selMsg == null) return;
		var notes = tbNotesDb.getNotesBySelectedMsg(selMsg);
		tbNotesUtils.removeChilds(hbox);
		
		if(notes.length > 0) {
			for(var n in notes) {
				var note = notes[n];
				var label = (note.mTitle == "") ? "..." : note.mTitle;
				var tip = tbNotes.mStrbundle.getFormattedString("editNoteLbl", [note.mTitle]);
				
				var button = document.createElementNS(XUL_NS, "button");
				button.setAttribute("label", label);
				button.setAttribute("tooltiptext", tip);
				button.setAttribute("type", "menu");
				button.setAttribute("class", "noteMsghdrButton");
				var menupopup = document.createElementNS(XUL_NS, "menupopup");
				var menuitem1 = document.createElementNS(XUL_NS, "menuitem");
				menuitem1.setAttribute("label", tbNotes.mStrbundle.getString("editNoteDlgTitle"));
				menuitem1.setAttribute("oncommand", "tbNotes.showEditNoteDialog('" + note.mId + "');");
				menupopup.appendChild(menuitem1);
				var menuseparator1 = document.createElementNS(XUL_NS, "menuseparator");
				menupopup.appendChild(menuseparator1);
				var menuitem2 = document.createElementNS(XUL_NS, "menuitem");
				menuitem2.setAttribute("label", tbNotes.mStrbundle.getString("createMailFromNoteLbl"));
				menuitem2.setAttribute("tooltiptext", tbNotes.mStrbundle.getString("createMailFromNoteTip"));
				menuitem2.setAttribute("oncommand", "tbNotes.createMailMessageFromNote('" + note.mId + "');");
				menuitem2.setAttribute("disabled", "true");
				menupopup.appendChild(menuitem2);
				var menuseparator2 = document.createElementNS(XUL_NS, "menuseparator");
				menupopup.appendChild(menuseparator2);
				var menuitem3 = document.createElementNS(XUL_NS, "menuitem");
				menuitem3.setAttribute("label", tbNotes.mStrbundle.getString("copyTitleLbl"));
				menuitem3.setAttribute("tooltiptext", tbNotes.mStrbundle.getString("copyTitleTip"));
				menuitem3.setAttribute("oncommand", "tbNotes.copyNoteTitleToClipboard('" + note.mId + "');");
				menupopup.appendChild(menuitem3);
				var menuitem4 = document.createElementNS(XUL_NS, "menuitem");
				menuitem4.setAttribute("label", tbNotes.mStrbundle.getString("copyTextLbl"));
				menuitem4.setAttribute("tooltiptext", tbNotes.mStrbundle.getString("copyTextTip"));
				menuitem4.setAttribute("oncommand", "tbNotes.copyNoteTextToClipboard('" + note.mId + "');");
				menupopup.appendChild(menuitem4);
				var menuseparator3 = document.createElementNS(XUL_NS, "menuseparator");
				menupopup.appendChild(menuseparator3);
				var menuitem5 = document.createElementNS(XUL_NS, "menuitem");
				menuitem5.setAttribute("label", tbNotes.mStrbundle.getString("removeNoteLbl"));
				menuitem5.setAttribute("oncommand", "tbNotes.deleteNote('" + note.mId + "');");
				menupopup.appendChild(menuitem5);
				button.appendChild(menupopup);
				hbox.appendChild(button);
			}
		}
		
		var button = document.createElementNS(XUL_NS, "button");
		button.setAttribute("image", "chrome://tbnotes-icons/content/tbnotes16.png");
		button.setAttribute("type", "menu");
		button.setAttribute("class", "noteMsghdrButtonIconic");
		var menupopup = document.createElementNS(XUL_NS, "menupopup");
		var menuitem1 = document.createElementNS(XUL_NS, "menuitem");
		menuitem1.setAttribute("label", tbNotes.mStrbundle.getString("addNoteLbl"));
		menuitem1.setAttribute("command", "cmd_addTbNote");
		menupopup.appendChild(menuitem1);
		
		if(notes.length > 0) {
			// this command is aplicable only when there are some notes created
			var menuseparator = document.createElementNS(XUL_NS, "menuseparator");
			menupopup.appendChild(menuseparator);
			var menuitem2 = document.createElementNS(XUL_NS, "menuitem");
			menuitem2.setAttribute("label", tbNotes.mStrbundle.getString("removeAllNotesLbl"));
			menuitem2.setAttribute("oncommand", "tbNotes.deleteNotes();");
			menupopup.appendChild(menuitem2);
		}
		
		button.appendChild(menupopup);
		hbox.appendChild(button);
	},
	
	onBeforeShowHeaderPane: function() {},
	onEndHeaders: function() {},
	onEndAttachements: function() {}
	
}; // End of tbNotesMessageListener

/**
 * Listener for nsIMsgFolderNotificationService. 
 * 
 * @type Object
 */
var tbNotesNotificationListener = {
	/**
	 * Notified after a command to delete a group of messages has been given, 
	 * but before the messages have actually been deleted.
	 * According to the current tbNotes preferences update notes attached 
	 * to any of messages which going to be deleted.
	 * 
	 * @param aMsgs array
	 * @returns {void}
	 */
	msgsDeleted: function(aMsgs)
	{
		//Components.utils.reportError("tbNotesNotificationListener.msgsDeleted");
		
		try {
			// `aMsgs` is an enumerator ...
			if(aMsgs.length > 0) {	
				var prefs = Components.classes["@mozilla.org/preferences-service;1"].
						getService(Components.interfaces.nsIPrefBranch);
				var pref = "extensions.tbnotes.how_act_when_deleting_msg";
				var what = prefs.getCharPref(pref);
				
				//Components.utils.reportError(what);
				
				for(let i=0; i<aMsgs.length; i++) {
					let msg = aMsgs.queryElementAt(i, Components.interfaces.nsIMsgDBHdr);
					
					//Components.utils.reportError(msg);
					//Components.utils.reportError(msg.messageKey);
					//Components.utils.reportError(msg.messageOffset);
					//Components.utils.reportError(msg.messageId);
					
					switch(what) {
						case "mark":
							// ...
							break;
						
						case "remove":
						default:
							// ...
							break;
					}
				}
			}
		} catch(e) {
			Components.utils.reportError(e);
		}
	},
	
	/**
	 * Fired when moving/copying of the selected messages is completed. We are
	 * interested only in moving.
	 *
	 * @param aMove {boolean}
	 * @param aSrcMsgs {array}
	 * @param aDestFolder {Components.interfaces.nsIMsgFolder}
	 * @returns {void}
	 */
	msgsMoveCopyCompleted: function(aMove, aSrcMsgs, aDestFolder) {
		//Components.utils.reportError("tbNotesNotificationListener.msgsMoveCopyCompleted");
		
		if(aMove == true) {
			
			for(var i=0; i<aSrcMsgs.length; i++) {
				//Components.utils.reportError(aSrcMsgs[i]);
			}
			
			//Components.utils.reportError(aDestFolder);
		}
	},
	
	folderDeleted: function(aFolder) {
		//Components.utils.reportError("tbNotesNotificationListener.folderDeleted");
		//Components.utils.reportError(aFolder);
		
		// TODO Search for all messages with note(s) attached?
	},
	
	// below are handlers for events that do not interest us
	folderMoveCopyCompleted: function(aMove, aSrcFolder, aDestFolder) {},
	folderRenamed: function(aOrigFolder, aNewFolder) {},
	msgAdded: function(aMsg) {},
	
	itemEvent: function(aItem, aEvent, aData) {
		//Components.utils.reportError("tbNotesNotificationListener.itemEvent");
		//Components.utils.reportError(aItem);
		//Components.utils.reportError(aEvent);
		//Components.utils.reportError(aData);
		
		// do nothing ...
	},
	
  QueryInterface: function(aIid) {
    // [ondrejd] - fix from tbNotes 0.3.1
    if(!aIid.equals(Components.interfaces.nsIMsgFolderListener) &&
	   /* previously CI.nsIMsgFolderNotificationServiceListener .... */
	   !aIid.equals(Components.interfaces.nsISupports)) {
      throw Components.results.NS_ERROR_NO_INTERFACE;
		}
		
    return this;
  }
	
}; // End of tbNotesNotificationListener

/**
 * Implementation of notes view tab
 *
 * @link https://bugzilla.mozila.org/attachment.cgi?id=378209&action=diff
 * @type Object
 * /
var tbNotesNotesViewTab = {
	name: "tbnotesTab",
	perTabPanel: "vbox",
	modes: {
		tbnotesTab: {
		  type: "tbnotesTab",
		  maxTabs: 1
		}
	},
	
	openTab: function tbnotesview_openTab(aTab, aArgs) {
		Components.utils.reportError("tbnotesview_openTab");
		
		aTab.title = tbNotes.mStrbundle.getString("notesViewTabName");
		
		let tab = document.getElementById("tbnotesTab");
		tab.setAttribute("collapsed", "false");
	},
	
	closeTab: function tbnotesview_closeTab(aTab) {
		let tab = document.getElementById("tbnotesTab");
		tab.setAttribute("collapsed", "true");
	},
	
	saveTab: function tbnotesview_saveTab(aTab) {},
	showTab: function tbnotesview_showTab(aTab) {
		Components.utils.reportError("tbnotesview_openTab");
		
		let tab = document.getElementById("tbnotesTab");
		tab.setAttribute("collapsed", "false");
		
	},
	
	shouldSwitchTo: function({ contentPage: aContentPage }) {
		let tabmail = document.getElementById("tabmail");
		let tabInfo = tabmail.tabInfo;

		for(let selectedIndex = 0; selectedIndex < tabInfo.length; ++selectedIndex) {
			if(tabInfo[selectedIndex].mode.name == "tbnotesTab") {
				return selectedIndex;
			}
		}
		return -1;
	},
	
	onTitleChanged: function tbnotesview_titleChanged(aTab) {},
	
	// TODO Implement this!
	supportsCommands: function tbnotesview_supportsCommand(aCommand, aTab) {
		Components.utils.reportError("tbnotesview_supportsCommand[aCommand]: " + aCommand);
		Components.utils.reportError("tbnotesview_supportsCommand[aTab]: " + aTab);
	},
	
	// TODO isCommandEnabled(aCommand, aTab)
	isCommandEnabled: function tbnotesview_isCommandEnabled(aCommand, aTab) {
		Components.utils.reportError("tbnotesview_isCommandEnabled[aCommand]: " + aCommand);
		Components.utils.reportError("tbnotesview_isCommandEnabled[aTab]: " + aTab);
	},
	
	// TODO doCommand(aCommand, aTab)
	doCommand: function tbnotesview_doCommand(aCommand, aTab) {
		Components.utils.reportError("tbnotesview_doCommand[aCommand]: " + aCommand);
		Components.utils.reportError("tbnotesview_doCommand[aTab]: " + aTab);
	}
	
};  // End of tbNotesNotesViewTab */

/**
 * Main object that contains all main functionality of tbNotes.
 * 
 * @type Object
 */
var tbNotes = {
	
	get mStrbundle() { return document.getElementById("tbnotes-stringbundle"); },
	
	/**
	 * Called when Thunderbird is loaded
	 * 
	 * @param aEvent {Components.interfaces.nsIDOMEvent}
	 * @returns {void}
	 */
	onLoad : function(aEvent) 
	{
		// Initialize tbNotes SQLite storage
		tbNotesDb.init();
		
		// Add our notes panel to the Thunderbird's tabmail
		//let tabmail = document.getElementById("tabmail");
		//tabmail.registerTabType(tbNotesNotesViewTab);
    
		// Add new col to the main Thunderbird's messages tree
		var obSrv = Components.classes["@mozilla.org/observer-service;1"].
				getService(Components.interfaces.nsIObserverService);
		obSrv.addObserver(tbNotesCreateDbObserver, "MsgCreateDBView", false);
		
		// Observe if any message is deleted
		var notificationSrv = Components.classes["@mozilla.org/messenger/msgnotificationservice;1"].
				getService(Components.interfaces.nsIMsgFolderNotificationService);
		notificationSrv.addListener(tbNotesNotificationListener,
		                            notificationSrv.msgAdded |
		                            notificationSrv.msgsDeleted |
		                            notificationSrv.msgsMoveCopyCompleted |
		                            // this code pre-dates msgsClassified
		                            // folderAdded intentionally omitted
		                            notificationSrv.folderDeleted |
		                            notificationSrv.folderMoveCopyCompleted |
		                            notificationSrv.folderRenamed);
		                            // itemEvent intentionally omitted); 
    
		// Add our listener to gMessagesListeners (defined in msgHdrViewOverlay.js)
		if(typeof(gMessageListeners) != "undefined") {
			gMessageListeners.push(tbNotesMessageListener);
		}
    
		// Add event listener for mail context menu
		var context = document.getElementById("mailContext");
		if(context) {
			context.addEventListener("popupshowing",
			                         tbNotes.onMailContextPopupShowing,
			                         false);
		}
	}, // end onLoad(aEvent)
	
	/**
	 * Fired when Thunderbird going to be quit
	 * 
	 * @param aEvent {Components.interfaces.nsIDOMEvent}
	 * @returns {void}
	 */
	onUnload : function(aEvent) 
	{
		try {
			var obSrv = Components.classes["@mozilla.org/observer-service;1"].
				getService(Components.interfaces.nsIObserverService);
			obSrv.removeObserver(tbNotesCreateDbObserver, "MsgCreateDBView", false);

			var notificationSrv = Components.classes["@mozilla.org/messenger/msgnotificationservice;1"].
				getService(Components.interfaces.nsIMsgFolderNotificationService);
			notificationSrv.removeListener(tbNotesNotificationListener); 

			var context = document.getElementById("mailContext");
			if(context) {
				context.removeEventListener("popupshowing", tbNotes.onMailContextPopupShowing, false);
			}
		} catch(e) {}
	}, // end onUnload(aEvent)
  
	/**
	 * Fired when mail context popupmenu is showing.
	 * 
	 * @param aEvent {Components.interfaces.nsIDOMEvent}
	 * @returns {void}
	 */
	onMailContextPopupShowing : function(aEvent) 
	{
		var menuitem1 = document.getElementById("tbn-addnote-menuitem");
		
		var selMsg = tbNotes.getSelectedMessageDescription();
		if(selMsg != null) {
			menuitem1.collapsed = false;
		} else {
			menuitem1.collapsed = true;
		}
	}, // end onMailContextPopupShowing(aEvent)
	
	/**
	 * Performs command 
	 * 
	 * @param aCmdName {string}
	 * @param aContext {string}
	 * @returns {void}
	 */
	doCommand : function(aCmdName, aContext) 
	{
		var selMsg = this.getSelectedMessageDescription();
		
		switch(aCmdName) {
			// adding new note
			case "add_note":        
				this.showAddNoteDialog(selMsg);
				break;
      
			// show notes view
			case "show_notesview":
				var tabmail = document.getElementById("tabmail");
				tabmail.openTab("contentTab", { contentPage: "chrome://tbnotes/content/notes-view/notes-view.xhtml" });
				// BUG: In SeaMonkey notes view doesn't work - we need the full tab implementation
				//tabmail.openTab("tbnotesTab");
				break;
      
			// show preferences dialog
			case "show_prefs":
				var dlg = window.openDialog("chrome://tbnotes/content/preferences/preferences.xul",
						"", "chrome,extrachrome,toolbar,resizable=yes,centerscreen,all");
				if(dlg) {
					dlg.focus();
				}
				break;
			
			// show about dialog
			case "show_about":
				var dlg = window.openDialog("chrome://tbnotes/content/about.xul", "",
						"chrome,extrachrome,dialog,resizable=no,centerscreen,all");
				if(dlg) {
					dlg.focus();
				}
				break;
			
			default:
				Components.utils.reportError("tbNotes.doCommand: Unknown command ['" + aCmdName + "']!");
				break;
		}
	}, // end doCommand(aCmdName, aContext)
	
	/**
	 * Returns simple object describing currently selected mail message or NULL.
	 * 
	 * @returns object|NULL
	 */
	getSelectedMessageDescription : function() 
	{
		if(gFolderDisplay.selectedMessage != null) {
			return {
				// BUG: When we're moving message with note attached the message key
				//      is changed when moved to another folder (or trash).
				key: gFolderDisplay.selectedMessage.messageKey,
				// We cann't use messageId because nsIMsgDBView doesn't contain method
				// similiar to the getKeyAt() - which is used to identify selected
				// message in the `tbNotesColumnHandler`
				//key: gFolderDisplay.selectedMessage.messageId,
				subject: gFolderDisplay.selectedMessage.subject,
				id: gFolderDisplay.selectedMessage.messageId
			};
		}
		
		return null;
	}, // end getSelectedMessageDescription()
	
	/**
	 * Returns TRUE if there is any note with given target message.
	 * 
	 * @param aMsg {object}
	 * @returns boolean
	 */
	checkNoteForTargetMsg : function(aMsg) 
	{
		return (tbNotesDb.getNoteBySelectedMsg(aMsg) == null) ? false : true;
	}, // end checkNoteForTargetMsg(aMsg)
	
	/**
	 * Show dialog for adding new note
	 *
	 * @param aSelMsg {object}
	 * @returns {void}
	 */
	showAddNoteDialog : function(aSelMsg) 
	{
		// Open dialog self
		var params = { inn: { msg: aSelMsg }, out: null };
		window.openDialog(
			"chrome://tbnotes/content/note-dialog.xul", "",
			"chrome,extrachrome,dialog,toolbar,modal,resizable=yes",
			params
		).focus();
		
		if(params.out != null) {
			params.out.save();
		}
		
		tbNotesMessageListener.onStartHeaders();
	}, // end showAddNoteDialog(aSelMsg)
	
	/** 
	 * Shows dialog for editing note
	 * 
	 * @param aNote {object|integer}
	 * @returns {void}
	 */
	showEditNoteDialog : function(aNote) 
	{
		if(typeof(aNote) != "object") {
			aNote = tbNotesDb.getNoteById(aNote);
		}
    
		var params = { inn: { note: aNote }, out: null };
		window.openDialog(
			"chrome://tbnotes/content/note-dialog.xul", "",
			"chrome,extrachrome,dialog,toolbar,modal,resizable=yes",
			params
		).focus();
		
		if(params.out != null) {
			params.out.save();
		}
		
		tbNotesMessageListener.onStartHeaders();
	}, // end showEditNoteDialog(aNote)
	
	/**
	 * Mark note as deleted
	 * 
	 * @param aNote {object|integer}
	 * @returns {void}
	 */
	deleteNote : function(aNote) 
	{
		if(typeof(aNote) != "object") {
			aNote = tbNotesDb.getNoteById(aNote);
		}
		
		if(typeof(aNote) != "object") {
			return;
		}
		
		var prefs = Components.classes["@mozilla.org/preferences-service;1"].
				getService(Components.interfaces.nsIPrefBranch);
		var pref = "extensions.tbnotes.dont_show_remove_confirm";
		
		// Check if user want to see confirmation dialog
		if(prefs.getBoolPref(pref) !== true) {
			// Show confirmation if user really want to close unsaved note
			var prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
					getService(Components.interfaces.nsIPromptService);
			var title = tbNotes.mStrbundle.getString("removeNoteDlgTitle");
			var desc = tbNotes.mStrbundle.getFormattedString("removeNoteDlgDesc", [aNote.mTitle]);
			var question = tbNotes.mStrbundle.getString("doNotAskAgain");
			var check = { value: false };
			var res = prompts.confirmCheck(window, title, desc, question, check);
			
			prefs.setBoolPref(pref, check.value);
			
			if(res) { 
				aNote.markDelete();
			}
		} else {
			aNote.markDelete();
		}
		
		tbNotesMessageListener.onStartHeaders();
	}, // end deleteNote(aNote)
	
	/**
	 * Delete all notes attached to the given message.
	 * 
	 * @returns {void}
	 */
	deleteNotes : function()
	{
		var selMsg = this.getSelectedMessageDescription();
		
		if(selMsg != null) {
			var notes = tbNotesDb.getNotesBySelectedMsg(selMsg);

			if(notes.length > 0) {
				for(var n in notes) {
					notes[n].markDelete();
				}
			}
		}
		
		tbNotesMessageListener.onStartHeaders();
	}, // end deleteNotes()
	
	/**
	 * Create mail message from the note
	 * 
	 * @param aNote {object|integer}
	 * @returns {boolean}
	 */
	createMailMessageFromNote: function(aNote)
	{
		if(typeof(aNote) != "object") {
			aNote = tbNotesDb.getNoteById(aNote);
		}
		
		// ...
		
	}, // end createMailMessageFromNote(aNote)
	
	/**
	 * Copy title of given note to the clipboard.
	 * 
	 * @param aNote {object|integer}
	 * @returns {boolean}
	 */
	copyNoteTitleToClipboard: function(aNote)
	{
		if(typeof(aNote) != "object") {
			aNote = tbNotesDb.getNoteById(aNote);
		}
		
		if(!aNote) return;
		
		tbNotesUtils.copyTextToClipboard(aNote.mTitle);
	}, // end copyNoteTitleToClipboard(aNote)
	
	/**
	 * Copy text of given note to the clipboard.
	 * 
	 * @param aNote {object|integer}
	 * @returns {void}
	 */
	copyNoteTextToClipboard: function(aNote)
	{
		if(typeof(aNote) != "object") {
			aNote = tbNotesDb.getNoteById(aNote);
		}
		
		if(!aNote) return;
		
		tbNotesUtils.copyTextToClipboard(aNote.mBody);
	} // end copyNoteTextToClipboard(aNote)
	
}; // End of tbNotes
